# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import io
import logging

from PIL import Image
from requests import Session
from requests.sessions import HTTPAdapter
from urllib3 import Retry

from datacube_ows.ogc_exceptions import WMSException

_LOG: logging.Logger = logging.getLogger(__name__)

RETRY_CODES = frozenset([413, 429, 500, 502, 503])

RETRIES = Retry(
    total=10,
    backoff_factor=0.5,
    raise_on_status=False,
    status_forcelist=RETRY_CODES,
)


def make_session(max_retries: Retry) -> Session:
    session = Session()
    session.mount("http://", HTTPAdapter(max_retries=max_retries))
    session.mount("https://", HTTPAdapter(max_retries=max_retries))
    return session


retrying_requests = make_session(RETRIES)


def get_image_from_url(url: str) -> Image.Image | None:
    """
    Fetch image a png from external URL, and return it as an Image.

    :param url:  A URL pointing to some png image
    :return: A PIL image object (OR None if the url does not return a PNG image)
    """
    r = retrying_requests.get(url)
    if r.status_code != 200:
        raise WMSException(
            f"Could not retrieve legend - external URL is failing with http code {r.status_code}"
        )
    if r.headers["content-type"] != "image/png":
        _LOG.warning(
            "External legend has MIME type %s. OWS strongly recommends PNG format for legend images.",
            r.headers["content-type"],
        )
    bytesio = io.BytesIO()
    bytesio.write(r.content)
    bytesio.seek(0)
    return Image.open(bytesio)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Iterable
from threading import Lock
from typing import Literal, cast
from uuid import UUID

import click
from datacube import Datacube
from datacube.model import Dataset, Product
from odc.geo import CRS, Geometry
from psycopg2.errors import ProgrammingError
from sqlalchemy import text
from typing_extensions import override

from datacube_ows.index.api import (
    InsufficientDbPrivileges,
    LayerExtent,
    LayerSignature,
    OWSAbstractIndex,
    OWSAbstractIndexDriver,
    TimeSearchTerm,
    check_perms,
)
from datacube_ows.index.sql import run_sql
from datacube_ows.ows_configuration import OWSNamedLayer

from .mv_index import MVSelectOpts, mv_search
from .product_ranges import create_range_entry as create_range_entry_impl
from .product_ranges import get_ranges as get_ranges_impl


class OWSPostgresIndex(OWSAbstractIndex):
    name: str = "postgres"

    # method to check database access (for ping op)
    @override
    def check_db_access(self, dc: Datacube) -> bool:
        db_ok = False
        try:
            with dc.index._db.give_me_a_connection() as conn:  # type: ignore[attr-defined]
                results = conn.execute(
                    text("""
                    SELECT *
                    FROM ows.layer_ranges
                    LIMIT 1""")
                )
                for _ in results:
                    db_ok = True
        except Exception:
            pass
        return db_ok

    @override
    def _check_perms(
        self, dc: Datacube, group: Literal["user", "manage", "admin"]
    ) -> None:
        try:
            with dc.index._db.give_me_a_connection() as conn:  # type: ignore[attr-defined]
                conn.execute(text(f"set role agdc_{group}"))
        except ProgrammingError as e:
            raise InsufficientDbPrivileges(
                f"db user {dc.index.environment.db_username} does not have agdc_{group} privileges: {e}"
            ) from None

    # method to delete obsolete schemas etc.
    @override
    @check_perms("admin")
    def cleanup_schema(self, dc: Datacube) -> None:
        self._run_sql(dc, "ows_schema/cleanup")

    # Schema creation method
    @override
    @check_perms("admin")
    def create_schema(self, dc: Datacube) -> None:
        click.echo("Creating schema and postgis extension...")
        self._run_sql(dc, "ows_schema/bootstrap")
        click.echo("Creating/updating tables...")
        self._run_sql(dc, "ows_schema/create")
        click.echo("Creating/updating materialised views...")
        self._run_sql(dc, "extent_views/create")
        click.echo("Granting tables permissions to agdc roles.")
        self._run_sql(dc, "ows_schema/grants")
        click.echo("Granting views permissions to agdc roles.")
        self._run_sql(dc, "extent_views/grants")

    # Spatiotemporal index update method (e.g. refresh materialised views)
    @override
    @check_perms("manage")
    def update_geotemporal_index(self, dc: Datacube) -> None:
        self._run_sql(dc, "extent_views/refresh")

    @override
    @check_perms("manage")
    def create_range_entry(
        self, layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]
    ) -> None:
        create_range_entry_impl(layer, cache)

    @override
    @check_perms("user")
    def get_ranges(self, layer: OWSNamedLayer) -> LayerExtent | None:
        return get_ranges_impl(layer)

    @override
    @check_perms("user")
    def ds_search(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[Dataset]:
        return cast(
            Iterable[Dataset],
            mv_search(
                layer.dc.index,
                MVSelectOpts.DATASETS,
                times=times,
                geom=geom,
                products=products,
            ),
        )

    @override
    @check_perms("user")
    def dsid_search(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[UUID]:
        return cast(
            Iterable[UUID],
            mv_search(
                layer.dc.index,
                MVSelectOpts.IDS,
                times=times,
                geom=geom,
                products=products,
            ),
        )

    @override
    @check_perms("user")
    def count(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> int:
        return cast(
            int,
            mv_search(
                layer.dc.index,
                MVSelectOpts.COUNT,
                times=times,
                geom=geom,
                products=products,
            ),
        )

    @override
    @check_perms("user")
    def extent(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
        crs: CRS | None = None,
    ) -> Geometry | None:
        extent = cast(
            Geometry | None,
            mv_search(
                layer.dc.index,
                MVSelectOpts.EXTENT,
                times=times,
                geom=geom,
                products=products,
            ),
        )
        if extent is None or crs is None or crs == extent.crs:
            return extent
        return extent.to_crs(crs)

    def _run_sql(self, dc: Datacube, path: str, **params: str) -> bool:
        return run_sql(dc, path, **params)


pgdriverlock = Lock()


class OWSPostgresIndexDriver(OWSAbstractIndexDriver):
    _driver = None

    @classmethod
    @override
    def ows_index_class(cls) -> type[OWSAbstractIndex]:
        return OWSPostgresIndex

    @classmethod
    @override
    def ows_index(cls) -> OWSAbstractIndex:
        with pgdriverlock:
            if cls._driver is None:
                cls._driver = OWSPostgresIndex()
        return cls._driver


def ows_index_driver_init() -> OWSPostgresIndexDriver:
    return OWSPostgresIndexDriver()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import math
from collections.abc import Callable
from datetime import date, datetime, timezone

import click
import datacube
import odc.geo
import sqlalchemy.exc
from odc.geo.crs import CRS
from psycopg.types.json import Json as Json3
from psycopg2.extras import Json as Json2
from sqlalchemy import text

from datacube_ows.index.api import CoordRange, LayerExtent, LayerSignature
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_driver_name, get_sqlconn

_LOG: logging.Logger = logging.getLogger(__name__)


def jsonise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    return {
        "top": bbox.top,
        "bottom": bbox.bottom,
        "left": bbox.left,
        "right": bbox.right,
    }


def create_range_entry(
    layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]
) -> None:
    # Take layer signature
    meta = LayerSignature(
        time_res=layer.time_resolution.value,
        products=tuple(layer.product_names),
        env=layer.local_env._name,
        datasets=layer.dc.index.datasets.count(product=layer.product_names),
    )

    conn = get_sqlconn(layer.dc)
    txn = conn.begin()
    if meta in cache:
        template = cache[meta][0]
        # Layer {template} has same signature - reusing
        cache[meta].append(layer.name)
        try:
            conn.execute(
                text("""
            INSERT INTO ows.layer_ranges
                (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
            SELECT :layer_id, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated
            FROM ows.layer_ranges lr2
            WHERE lr2.layer = :template_id"""),
                {"layer_id": layer.name, "template_id": template},
            )
        except sqlalchemy.exc.IntegrityError:
            conn.execute(
                text("""
            UPDATE ows.layer_ranges lr1
            SET lat_min = lr2.lat_min,
                lat_max = lr2.lat_max,
                lon_min = lr2.lon_min,
                lon_max = lr2.lon_max,
                dates = lr2.dates,
                bboxes = lr2.bboxes,
                meta = lr2.meta,
                last_updated = lr2.last_updated
            FROM ows.layer_ranges lr2
            WHERE lr1.layer = :layer_id
            AND   lr2.layer = :template_id"""),
                {"layer_id": layer.name, "template_id": template},
            )
    else:
        Json = Json3 if get_driver_name(layer.dc.index) == "psycopg" else Json2
        # insert empty row if one does not already exist
        conn.execute(
            text("""
        INSERT INTO ows.layer_ranges
        (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
        VALUES
        (:p_layer, 0, 0, 0, 0, :empty, :empty, :meta, :now)
        ON CONFLICT (layer) DO NOTHING
        """),
            {
                "p_layer": layer.name,
                "empty": Json(""),
                "meta": Json(meta.as_json()),
                "now": datetime.now(tz=timezone.utc),
            },
        )

        prodids = [p.id for p in layer.products]

        # Set default timezone
        conn.execute(text("""set timezone to 'Etc/UTC'"""))

        # Loop over dates
        dates = set()  # Should get to here!
        if layer.time_resolution.is_solar():
            results = conn.execute(
                text(
                    """
              select lower(dt.search_val), upper(dt.search_val),
                     lower(lon.search_val), upper(lon.search_val)
              from odc.dataset ds, odc.dataset_search_datetime dt, odc.dataset_search_num lon
              where ds.product_ref = ANY(:prodids)
              AND ds.id = dt.dataset_ref
              AND ds.id = lon.dataset_ref
              AND dt.search_key = :time
              AND lon.search_key = :lon
              """
                ),
                {"prodids": prodids, "time": "time", "lon": "lon"},
            )
            for result in results:
                dt1, dt2, ll, lu = result
                lon = (ll + lu) / 2
                dt = dt1 + (dt2 - dt1) / 2
                dt = dt.astimezone(timezone.utc)

                solar_day = datacube.api.query._convert_to_solar_time(dt, lon).date()
                dates.add(solar_day)
        else:
            results = conn.execute(
                text(
                    """
              select
                    array_agg(dt.search_val)
              from odc.dataset_search_datetime dt,
                   odc.dataset ds
              WHERE ds.product_ref = ANY(:prodids)
              AND   ds.id = dt.dataset_ref
              AND   dt.search_key = 'time'
              """
                ),
                {"prodids": prodids},
            )
            for result in results:
                for dat_ran in result[0]:
                    dates.add(dat_ran.lower)

        if layer.time_resolution.is_subday():
            date_formatter = lambda d: d.isoformat()  # noqa: E731
        else:
            date_formatter = lambda d: d.strftime("%Y-%m-%d")  # noqa: E731

        dates = sorted(dates)
        conn.execute(
            text("""
           UPDATE ows.layer_ranges
           SET dates = :dates
           WHERE layer= :layer_id
        """),
            {"dates": Json(list(map(date_formatter, dates))), "layer_id": layer.name},
        )
        # calculate bounding boxes
        # Get extent polygon from materialised views

        all_bboxes = build_bboxes(layer)
        if not all_bboxes:
            return

        conn.execute(
            text("""
        UPDATE ows.layer_ranges
        SET bboxes = :bbox,
            lat_min = :lat_min,
            lat_max = :lat_max,
            lon_min = :lon_min,
            lon_max = :lon_max
        WHERE layer = :layer_id
        """),
            {
                "bbox": Json(all_bboxes),
                "layer_id": layer.name,
                "lat_min": all_bboxes["EPSG:4326"]["bottom"],
                "lat_max": all_bboxes["EPSG:4326"]["top"],
                "lon_min": all_bboxes["EPSG:4326"]["left"],
                "lon_max": all_bboxes["EPSG:4326"]["right"],
            },
        )

        cache[meta] = [layer.name]

    txn.commit()
    conn.close()


def build_bboxes(layer: OWSNamedLayer) -> dict[str, dict[str, float]]:
    # User native CRS for base extent, unless no spatial index for it, in which case
    # use 4326 which always has a spatial index.
    base_crs = CRS(layer.native_CRS)
    if base_crs not in layer.dc.index.spatial_indexes():
        base_crs = CRS("epsg:4326")
    base_extent = extent_for_layer(layer, base_crs)
    if base_extent is None:
        click.echo(f"Layer {layer.name} has no extent in CRS {base_crs}.  Skipping.")
        return {}
    result = {}
    for crsid, crs in layer.global_config().crses.items():
        if crs == base_crs:
            result[crsid] = sanitise_bbox(base_extent.boundingbox)
        elif crs in layer.dc.index.spatial_indexes():
            extent = extent_for_layer(layer, crs)
            if extent is None:
                click.echo(f"Layer {layer.name} has no extent in CRS {crs}.  Skipping.")
                return {}
            result[crsid] = sanitise_bbox(extent.boundingbox)
        else:
            extent = base_extent.to_crs(crs)
            result[crsid] = sanitise_bbox(extent.boundingbox)
    return result


def extent_for_layer(layer: OWSNamedLayer, crs: CRS) -> odc.geo.Geometry | None:
    ext = None
    for product in layer.products:
        prod_extent = layer.dc.index.products.spatial_extent(product, crs)
        ext = prod_extent if ext is None else ext | prod_extent
    return ext


def bbox_projections(
    starting_box: odc.geo.Geometry, crses: dict[str, odc.geo.CRS]
) -> dict[str, dict[str, float]]:
    result = {}
    for crsid, crs in crses.items():
        projbbox = starting_box.to_crs(crs).boundingbox
        result[crsid] = sanitise_bbox(projbbox)
    return result


def sanitise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    def sanitise_coordinate(coord: float, fallback: float, upper: bool) -> float:
        if (
            not math.isfinite(coord)
            or (upper and coord > fallback)
            or (not upper and coord < fallback)
        ):
            return fallback
        return coord

    if bbox.crs == CRS("epsg:4326"):
        return {
            "top": sanitise_coordinate(bbox.top, float("90.0"), True),
            "bottom": sanitise_coordinate(bbox.bottom, float("-90.0"), False),
            "left": sanitise_coordinate(bbox.left, float("-180"), False),
            "right": sanitise_coordinate(bbox.right, float("9.999999999e99"), True),
        }
    return {
        "top": sanitise_coordinate(bbox.top, float("9.999999999e99"), True),
        "bottom": sanitise_coordinate(bbox.bottom, float("-9.999999999e99"), False),
        "left": sanitise_coordinate(bbox.left, float("-9.999999999e99"), False),
        "right": sanitise_coordinate(bbox.right, float("9.999999999e99"), True),
    }


def get_ranges(layer: OWSNamedLayer) -> LayerExtent | None:
    conn = get_sqlconn(layer.dc)
    results = conn.execute(
        text("""
        SELECT *
        FROM ows.layer_ranges
        WHERE layer=:pname"""),
        {"pname": layer.name},
    )

    for result in results:
        conn.close()
        if layer.time_resolution.is_subday():
            dt_parser: Callable[[str], datetime | date] = (  # noqa: E731
                lambda dts: datetime.fromisoformat(dts)
            )
        else:
            dt_parser = lambda dts: datetime.strptime(dts, "%Y-%m-%d").date()  # noqa: E731
        times = [dt_parser(d) for d in result.dates if d is not None]
        if not times:
            return None
        return LayerExtent(
            lat=CoordRange(min=float(result.lat_min), max=float(result.lat_max)),
            lon=CoordRange(min=float(result.lon_min), max=float(result.lon_max)),
            times=times,
            bboxes=result.bboxes,
        )
    return None



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from datetime import date, datetime
from typing import Any, Literal, NamedTuple, TypeAlias, Union
from uuid import UUID

from datacube import Datacube
from datacube.index.abstract import AbstractIndex
from datacube.model import Dataset, Product
from odc.geo.crs import CRS
from odc.geo.geom import Geometry, polygon

from datacube_ows.config_utils import CFG_DICT, ConfigException

TYPE_CHECKING = False
if TYPE_CHECKING:
    from datacube_ows.ows_configuration import OWSNamedLayer


class AbortRun(Exception):
    pass


class InsufficientDbPrivileges(AbortRun):
    pass


@dataclasses.dataclass(frozen=True)
class LayerSignature:
    time_res: str
    products: tuple[str, ...]
    env: str
    datasets: int

    def as_json(self) -> dict[str, list[str] | str | int]:
        return {
            "time_res": self.time_res,
            "products": list(self.products),
            "env": self.env,
            "datasets": self.datasets,
        }


DateOrDateTime: TypeAlias = datetime | date
TimeSearchTerm: TypeAlias = (
    tuple[datetime, datetime] | tuple[date, date] | DateOrDateTime
)


class CoordRange(NamedTuple):
    min: float
    max: float


class LayerExtent:
    def __init__(
        self,
        lat: CoordRange,
        lon: CoordRange,
        times: list[DateOrDateTime],
        bboxes: CFG_DICT,
    ) -> None:
        self.lat = lat
        self.lon = lon
        self.times = times
        self.start_time = times[0]
        self.end_time = times[-1]
        self.time_set = set(times)
        self.bboxes = bboxes


class OWSAbstractIndex(ABC):
    name: str = ""

    # method to check if we are in a user group
    # N.B. Subclasses may add additional groups to the Literal as required.
    @abstractmethod
    def _check_perms(self, dc: Datacube, group: Literal["admin"]) -> None: ...

    # method to check database access (for ping op) (requires odc "user" perms)
    @abstractmethod
    def check_db_access(self, dc: Datacube) -> bool: ...

    # method to delete obsolete schemas etc. (requires odc "admin" perms)
    @abstractmethod
    def cleanup_schema(self, dc: Datacube): ...

    # Schema creation method (requires odc "admin" perms)
    @abstractmethod
    def create_schema(self, dc: Datacube): ...

    # Spatiotemporal index update method (e.g. refresh materialised views) (requires odc "manage" perms)
    @abstractmethod
    def update_geotemporal_index(self, dc: Datacube): ...

    # Range table update method (requires odc "manage" perms)
    @abstractmethod
    def create_range_entry(
        self, layer: "OWSNamedLayer", cache: dict[LayerSignature, list[str]]
    ) -> None: ...

    # Range table read method (requires odc "user" perms)
    @abstractmethod
    def get_ranges(self, layer: "OWSNamedLayer") -> LayerExtent | None: ...

    # Spatiotemporal search methods (requires odc "user" perms)
    @abstractmethod
    def ds_search(
        self,
        layer: "OWSNamedLayer",
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[Dataset]: ...

    def dsid_search(
        self,
        layer: "OWSNamedLayer",
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[UUID]:
        for ds in self.ds_search(layer, times, geom, products):
            yield ds.id

    def count(
        self,
        layer: "OWSNamedLayer",
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> int:
        return len(list(self.dsid_search(layer, times, geom, products)))

    def extent(
        self,
        layer: "OWSNamedLayer",
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
        crs: CRS | None = None,
    ) -> Geometry | None:
        geom = self._prep_geom(layer, geom)
        if crs is None:
            crs = CRS("epsg:4326")
        ext: Geometry | None = None
        # Accumulate extent in native CRS if possible.
        for ds in self.ds_search(layer, times, geom, products):
            if ds.extent:
                if ds.extent.crs != CRS(layer.native_CRS):
                    # Reproject to layer "native" CRS if needed.
                    ds_extent: Geometry = ds.extent.to_crs(layer.native_CRS)
                else:
                    ds_extent = ds.extent
                ext = ds_extent if ext is None else ext.union(ds_extent)
        if ext is not None and crs != CRS(layer.native_CRS):
            # Reproject to requested CRS if necessary
            return ext.to_crs(crs)
        return ext

    @staticmethod
    def _prep_geom(
        layer: "OWSNamedLayer", any_geom: Geometry | None
    ) -> Geometry | None:
        # Prepare a Geometry for geospatial search
        # Perhaps Core can be updated so this is not needed?
        if any_geom is None:
            # None?  Leave as None
            return None
        if any_geom.geom_type == "Point":
            # Point?  Expand to a polygon covering a single native pixel.
            any_geom = any_geom.to_crs(layer.native_CRS)
            x, y = any_geom.coords[0]
            delta_x, delta_y = layer.cfg_native_resolution
            return polygon(
                (
                    (x, y),
                    (x + delta_x, y),
                    (x + delta_x, y + delta_y),
                    (x, y + delta_y),
                    (x, y),
                ),
                crs=layer.native_CRS,
            )
        if any_geom.geom_type in ("MultiPoint", "LineString", "MultiLineString"):
            # Not a point, but not a polygon or multipolygon?  Expand to polygon by taking convex hull
            return any_geom.convex_hull
        # Return polygons and multipolygons as is.
        return any_geom


class OWSAbstractIndexDriver(ABC):
    @classmethod
    @abstractmethod
    def ows_index_class(cls) -> type[OWSAbstractIndex]: ...

    @classmethod
    @abstractmethod
    def ows_index(cls) -> OWSAbstractIndex: ...


def ows_index(odc: Datacube | AbstractIndex) -> OWSAbstractIndex:
    index = odc if isinstance(odc, AbstractIndex) else odc.index
    env = index.environment
    from datacube_ows.index.driver import ows_index_driver_by_name

    idx_drv_name = (
        "postgres" if env.index_driver in ("default", "legacy") else env.index_driver
    )
    ows_index_driver = ows_index_driver_by_name(idx_drv_name)
    if ows_index_driver is None:
        raise ConfigException(
            f"ODC Environment {env._name} uses ODC index driver {env.index_driver} which is "
            "not (yet) supported by OWS."
        )
    return ows_index_driver.ows_index()


def check_perms(
    group: Literal["user", "admin", "manage"],
) -> Callable[[Callable], Callable]:
    def outer(f: Callable) -> Callable:
        def inner(
            instance, dcl: Union[Datacube, "OWSNamedLayer"], *args, **kwargs
        ) -> Any:
            from datacube_ows.ows_configuration import OWSNamedLayer

            if isinstance(dcl, OWSNamedLayer):
                dc = Datacube(env=dcl.local_env)
            elif isinstance(dcl, Datacube):
                dc = dcl
            else:
                raise TypeError("Expected Datacube or OWSNamedLayer")
            instance._check_perms(dc, group)
            return f(instance, dcl, *args, **kwargs)

        return inner

    return outer



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import importlib
import re

import click
import datacube.cfg
import sqlalchemy
from datacube import Datacube
from psycopg2.errors import DuplicateObject, InsufficientPrivilege

from datacube_ows.index.api import InsufficientDbPrivileges


def get_sqlconn(dc: Datacube) -> sqlalchemy.Connection:
    """
    Extracts a SQLAlchemy database connection from a Datacube object.

    :param dc: An initialised Datacube object
    :return: A SQLAlchemy database connection object.
    """
    # pylint: disable=protected-access
    return dc.index._db._engine.connect()  # type: ignore[attr-defined]


def run_sql(dc: Datacube, path: str, **params: str) -> bool:
    driver_names = {
        "pg_index": "postgres",
        "pgis_index": "postgis",
    }
    driver_name = driver_names[dc.index.name]
    print(f"path in is {path}")
    full_path = importlib.resources.files("datacube_ows").joinpath(
        f"sql/{driver_name}/{path}"
    )
    if not full_path.is_dir():
        print(
            f"Cannot find SQL resource directory {full_path} - check your datacube-ows installation"
        )
        return False

    files = sorted(full_path.iterdir())  # type: ignore[type-var]

    # N.B. We aren't actually using this "required parameters" feature at
    #      the moment.
    filename_req_pattern = re.compile(
        r"\d+[_a-zA-Z0-9]+_requires_(?P<reqs>[_a-zA-Z0-9]+)\.sql"
    )
    filename_pattern = re.compile(r"\d+[_a-zA-Z0-9]+\.sql")
    with get_sqlconn(dc) as conn:
        all_ok: bool = True
        for fi in files:
            fname = fi.name
            isolated = fname.endswith("_isolated.sql")
            match = filename_pattern.fullmatch(fname)
            if not match:
                click.echo(f"Illegal SQL filename: {fname} (skipping)")
                all_ok = False
                continue
            req_match = filename_req_pattern.fullmatch(fname)
            reqs = req_match.group("reqs").split("_") if req_match else []
            if reqs:
                try:
                    kwargs = {v: params[v] for v in reqs if v != "isolated"}
                except KeyError as e:
                    click.echo(
                        f"Required parameter {e} for file {fname} not supplied - skipping"
                    )
                    all_ok = False
                    continue
            else:
                kwargs = {}
            comment, sql = read_file(driver_name, path, fname, **kwargs)
            if reqs:
                sql = sql.format(**kwargs)
            if isolated:
                conn.commit()
                with get_sqlconn(dc).execution_options(
                    isolation_level="AUTOCOMMIT"
                ) as iso_conn:
                    run_sql_statement(
                        sql, comment, fname, iso_conn, dc.index.environment
                    )
            else:
                run_sql_statement(sql, comment, fname, conn, dc.index.environment)

        return all_ok


def read_file(
    driver_name: str, path: str, fname: str, **kwargs: str
) -> tuple[str, str]:
    ref = importlib.resources.files("datacube_ows").joinpath(
        f"sql/{driver_name}/{path}/{fname}"
    )
    comment = ""
    sql = ""
    with ref.open("rb") as fp:
        first = True
        for line in fp:
            sline = str(line, "utf-8")
            if first and sline.startswith("--"):
                comment = sline[2:].format(**kwargs) if kwargs else sline[2:]
            else:
                sql = sql + "\n" + sline
            first = False
    return comment, sql


def run_sql_statement(
    sql: str,
    comment: str,
    fname: str,
    conn: sqlalchemy.Connection,
    env: datacube.cfg.ODCEnvironment,
) -> None:
    click.echo(f" - Running SQL statement: {comment}")
    try:
        result = conn.execute(sqlalchemy.text(sql))
        click.echo(f"    ...  succeeded(?) with rowcount {result.rowcount}")
    except sqlalchemy.exc.ProgrammingError as e:
        if isinstance(e.orig, InsufficientPrivilege):
            raise InsufficientDbPrivileges(
                f"Insufficient Privileges (user {env.db_username}). Try running again as a database superuser"
            ) from None
        if isinstance(e.orig, DuplicateObject):
            if fname.endswith("_ignore_duplicates.sql"):
                click.echo("Ignoring 'already exists' error")
            else:
                click.echo(f"Unexpected database error: {e}")
                raise e from None
        else:
            click.echo(f"Unexpected database error: {e}")
            raise e from e



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any

from flask import request
from ows.common import WGS84BoundingBox
from ows.common.v20.decoders import kvp_decode_get_capabilities
from ows.gml import Grid, IrregularAxis, RegularAxis, SpatioTemporalType
from ows.swe import Field
from ows.wcs import CoverageDescription, CoverageSummary, ServiceCapabilities
from ows.wcs.v20 import encoders as encoders_v20
from ows.wcs.v20.decoders import kvp_decode_describe_coverage, kvp_decode_get_coverage
from ows.wcs.v21 import encoders as encoders_v21

from datacube_ows.http_utils import (
    cache_control_headers,
    get_service_base_url,
    json_response,
    resp_headers,
)
from datacube_ows.ogc_exceptions import WCS2Exception
from datacube_ows.ows_configuration import OWSConfig, get_config
from datacube_ows.query_profiler import QueryProfiler
from datacube_ows.utils import log_call
from datacube_ows.wcs2_utils import get_coverage_data

_LOG: logging.Logger = logging.getLogger(__name__)

WCS_REQUESTS = ("DESCRIBECOVERAGE", "GETCOVERAGE")


@log_call
def handle_wcs2(nocase_args) -> tuple:
    operation = nocase_args.get("request", "").upper()
    if not operation:
        raise WCS2Exception("No operation specified", locator="Request parameter")
    if operation == "GETCAPABILITIES":
        return get_capabilities(nocase_args)
    if operation == "DESCRIBECOVERAGE":
        return desc_coverages(nocase_args)
    if operation == "GETCOVERAGE":
        return get_coverage(
            request.args.lists(),
            bool(nocase_args.get("ows_stats")),
            nocase_args.get("styles"),
        )
    raise WCS2Exception(
        f"Unrecognised operation: {operation}", locator="Request parameter"
    )


@log_call
def get_capabilities(args: dict) -> tuple[Any, int, dict]:
    # Extract layer metadata from Datacube.
    cfg = get_config()
    url = args.get("Host", args["url_root"])
    base_url = get_service_base_url(cfg.allowed_urls, url)

    request_obj = kvp_decode_get_capabilities(args)
    sections = request_obj.sections or ["all"]

    # TODO: check for invalid sections
    include_service_identification = False
    include_service_provider = False
    include_operations_metadata = False
    include_service_metadata = False
    include_coverage_summary = False

    if "all" in sections:
        include_service_identification = True
        include_service_provider = True
        include_operations_metadata = True
        include_service_metadata = True
        include_coverage_summary = True
    if "serviceidentification" in sections:
        include_service_identification = True
    if "serviceprovider" in sections:
        include_service_provider = True
    if "operationsmetadata" in sections:
        include_operations_metadata = True
    if "servicemetadata" in sections:
        include_service_metadata = True
    if "coveragesummary" in sections:
        include_coverage_summary = True

    assert cfg.contact_info is not None
    assert cfg.contact_info.address is not None
    capabilities = ServiceCapabilities.with_defaults_v20(
        service_url=base_url + "/wcs",
        allowed_operations=["GetCapabilities", "DescribeCoverage", "GetCoverage"],
        allow_post=False,
        title=cfg.title,
        abstract=cfg.abstract,
        keywords=cfg.keywords,
        fees=cfg.fees,
        access_constraints=[cfg.access_constraints],
        provider_name="",
        provider_site="",
        individual_name=cfg.contact_info.person,
        organisation_name=cfg.contact_info.organisation,
        position_name=cfg.contact_info.position,
        phone_voice=cfg.contact_info.telephone,
        phone_facsimile=cfg.contact_info.fax,
        delivery_point=cfg.contact_info.address.address,
        city=cfg.contact_info.address.city,
        administrative_area=cfg.contact_info.address.state,
        postal_code=cfg.contact_info.address.postcode,
        country=cfg.contact_info.address.country,
        electronic_mail_address=cfg.contact_info.email,
        online_resource=base_url,
        # hours_of_service=,
        # contact_instructions=,
        # role=,
        coverage_summaries=[
            CoverageSummary(
                identifier=product.name,
                coverage_subtype="RectifiedGridCoverage",
                title=product.title,
                wgs84_bbox=WGS84BoundingBox(
                    [
                        product.ranges.lon.min,
                        product.ranges.lat.min,
                        product.ranges.lon.max,
                        product.ranges.lat.max,
                    ]
                ),
            )
            for product in cfg.layer_index.values()
            if product.ready and not product.hide and product.wcs
        ],
        formats_supported=[fmt.mime for fmt in cfg.wcs_formats if 2 in fmt.renderers],
        crss_supported=list(cfg.published_CRSs),  # TODO: conversion to URL format
        interpolations_supported=None,  # TODO: find out interpolations
    )
    result = encoders_v20.xml_encode_capabilities(
        capabilities,
        include_service_identification=include_service_identification,
        include_service_provider=include_service_provider,
        include_operations_metadata=include_operations_metadata,
        include_service_metadata=include_service_metadata,
        include_coverage_summary=include_coverage_summary,
    )

    headers = cache_control_headers(cfg.wms_cap_cache_age)
    headers["Content-Type"] = result.content_type
    return (result.value, 200, cfg.response_headers(headers))


def create_coverage_description(cfg: OWSConfig, product) -> CoverageDescription:
    axes = [
        RegularAxis(
            label=product.native_CRS_def["horizontal_coord"],
            index_label="i",
            lower_bound=min(
                product.ranges.bboxes[product.native_CRS]["left"],
                product.ranges.bboxes[product.native_CRS]["right"],
            ),
            upper_bound=max(
                product.ranges.bboxes[product.native_CRS]["left"],
                product.ranges.bboxes[product.native_CRS]["right"],
            ),
            resolution=product.resolution_x,
            uom="deg",
            size=product.grid_high_x,
        ),
        RegularAxis(
            label=product.native_CRS_def["vertical_coord"],
            index_label="j",
            lower_bound=min(
                product.ranges.bboxes[product.native_CRS]["top"],
                product.ranges.bboxes[product.native_CRS]["bottom"],
            ),
            upper_bound=max(
                product.ranges.bboxes[product.native_CRS]["top"],
                product.ranges.bboxes[product.native_CRS]["bottom"],
            ),
            resolution=product.resolution_y,
            uom="deg",
            size=product.grid_high_y,
        ),
    ]

    # swap axes if necessary
    if product.native_CRS_def.get("vertical_coord_first"):
        axes.reverse()

    if product.regular_time_axis:
        start, end = product.time_range()
        size = (end - start).days // product.time_axis_interval
        axes.append(
            RegularAxis(
                label="time",
                index_label="k",
                lower_bound=start.isoformat(),
                upper_bound=end.isoformat(),
                resolution=f"P{product.time_axis_interval}D",
                uom="ISO-8601",
                type=SpatioTemporalType.TEMPORAL,
                size=size,
            )
        )
    elif product.time_resolution.is_subday():
        axes.append(
            IrregularAxis(
                label="time",
                index_label="k",
                positions=[f"{t.isoformat()}" for t in product.ranges.times],
                uom="ISO-8601",
                type=SpatioTemporalType.TEMPORAL,
            )
        )
    else:
        axes.append(
            IrregularAxis(
                label="time",
                index_label="k",
                positions=[
                    f"{t.isoformat()}T00:00:00.000Z" for t in product.ranges.times
                ],
                uom="ISO-8601",
                type=SpatioTemporalType.TEMPORAL,
            )
        )

    return CoverageDescription(
        identifier=product.name,
        title=product.title,
        abstract=product.definition.get("description"),
        range_type=[
            Field(
                name=band_label,
                description=band_label,
                uom=band_label,
                nil_values=dict.fromkeys(
                    product.band_idx.band_nodata_vals(), "invalid"
                ),
            )
            for band_label in product.band_idx.band_labels()
        ],
        grid=Grid(axes=axes, srs=product.native_CRS),
        native_format=cfg.wcs_formats_by_name[product.native_format].mime,
        coverage_subtype="RectifiedGridCoverage",
    )


@log_call
def desc_coverages(args) -> tuple:
    cfg = get_config()

    request_obj = kvp_decode_describe_coverage(args)

    products = []

    for coverage_id in request_obj.coverage_ids:
        product = cfg.layer_index.get(coverage_id)
        if product and product.wcs:
            products.append(product)
        else:
            raise WCS2Exception(
                f"Invalid coverage: {coverage_id}",
                WCS2Exception.NO_SUCH_COVERAGE,
                locator=coverage_id,
            )

    coverage_descriptions = [
        create_coverage_description(cfg, product) for product in products
    ]

    version = request_obj.version

    if version == (2, 0):
        result = encoders_v20.xml_encode_coverage_descriptions(coverage_descriptions)
    elif version == (2, 1):
        result = encoders_v21.xml_encode_coverage_descriptions(coverage_descriptions)
    else:
        raise WCS2Exception(
            f"Unsupported version: {version}",
            WCS2Exception.INVALID_PARAMETER_VALUE,
            locator="version",
        )
    min_cache_age = min(p.resource_limits.wcs_desc_cache_rule for p in products)
    headers = cache_control_headers(min_cache_age)
    headers["Content-Type"] = result.content_type
    return (result.value, 200, resp_headers(headers))


@log_call
def get_coverage(
    args, ows_stats: bool = False, styles=None
) -> tuple[str | bytes, int, dict[str, str]]:
    request_obj = kvp_decode_get_coverage(args)
    qprof = QueryProfiler(ows_stats)
    output, headers = get_coverage_data(request_obj, styles, qprof)
    if ows_stats:
        return json_response(qprof.profile())
    return (output, 200, resp_headers(headers))



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import collections
import logging
from datetime import date, datetime

import numpy
import xarray as xr
from dateutil.parser import parse
from odc.geo.geobox import GeoBox
from ows.wcs.v20 import ScaleAxis, ScaleExtent, ScaleSize, Slice, Trim
from rasterio import MemoryFile

from datacube_ows.loading import DataStacker
from datacube_ows.ogc_exceptions import WCS2Exception
from datacube_ows.ows_configuration import OWSConfig, get_config
from datacube_ows.resource_limits import ResourceLimited
from datacube_ows.utils import default_to_utc
from datacube_ows.wcs_scaler import WCSScaler, WCSScalerUnknownDimension

# from datacube_ows.wcs_utils import get_bands_from_styles

_LOG: logging.Logger = logging.getLogger(__name__)


def uniform_crs(cfg: OWSConfig, crs: str) -> str:
    """Helper function to transform a URL style EPSG definition to an 'EPSG:nnn' one"""
    if crs.startswith("http://www.opengis.net/def/crs/EPSG/"):
        code = crs.rpartition("/")[-1]
        crs = f"EPSG:{code}"
    elif crs.startswith("urn:ogc:def:crs:EPSG:"):
        code = crs.rpartition(":")[-1]
        crs = f"EPSG:{code}"
    elif crs.startswith("EPSG") or crs in cfg.published_CRSs:
        pass
    else:
        raise WCS2Exception(
            f"Not a CRS: {crs}",
            WCS2Exception.NOT_A_CRS,
            locator=crs,
            valid_keys=list(cfg.published_CRSs),
        )
    return crs


def get_coverage_data(request, styles, qprof) -> tuple:
    # pylint: disable=too-many-locals, protected-access

    cfg = get_config()
    qprof.start_event("setup")
    layer_name = request.coverage_id
    layer = cfg.layer_index.get(layer_name)
    if not layer or not layer.wcs:
        raise WCS2Exception(
            f"Invalid coverage: {layer_name}",
            WCS2Exception.NO_SUCH_COVERAGE,
            locator="COVERAGE parameter",
            valid_keys=list(cfg.layer_index),
        )

    #
    # CRS handling
    #

    native_crs = layer.native_CRS
    subsetting_crs = uniform_crs(cfg, request.subsetting_crs or native_crs)
    if subsetting_crs not in cfg.published_CRSs:
        raise WCS2Exception(
            f"Invalid subsettingCrs: {subsetting_crs}",
            WCS2Exception.SUBSETTING_CRS_NOT_SUPPORTED,
            locator=subsetting_crs,
            valid_keys=list(cfg.published_CRSs),
        )

    output_crs = uniform_crs(cfg, request.output_crs or subsetting_crs or native_crs)

    if output_crs not in cfg.published_CRSs:
        raise WCS2Exception(
            f"Invalid outputCrs: {output_crs}",
            WCS2Exception.OUTPUT_CRS_NOT_SUPPORTED,
            locator=output_crs,
            valid_keys=list(cfg.published_CRSs),
        )

    #
    # Subsetting/Scaling
    #

    scaler = WCSScaler(layer, subsetting_crs)
    times = layer.ranges.times

    subsets = request.subsets

    if len(subsets) != len({subset.dimension.lower() for subset in subsets}):
        dimensions = [subset.dimension.lower() for subset in subsets]
        duplicate_dimensions = [
            item for item, count in collections.Counter(dimensions).items() if count > 1
        ]

        raise WCS2Exception(
            f"Duplicate dimension{'s' if len(duplicate_dimensions) > 1 else ''}: "
            f"{', '.join(duplicate_dimensions)}",
            WCS2Exception.INVALID_SUBSETTING,
            locator=",".join(duplicate_dimensions),
        )

    for subset in subsets:
        dimension = subset.dimension.lower()
        if dimension == "time":
            if isinstance(subset, Trim):
                if "," in subset.high:
                    raise WCS2Exception(
                        "Subsets can only contain 2 elements - the lower and upper bounds. For arbitrary date lists, use WCS1",
                        WCS2Exception.INVALID_SUBSETTING,
                        locator="time",
                    )
                if layer.time_resolution.is_subday():
                    low: date | datetime | None = (
                        default_to_utc(parse(subset.low))
                        if subset.low is not None
                        else None
                    )
                    high: date | datetime | None = (
                        default_to_utc(parse(subset.high))
                        if subset.high is not None
                        else None
                    )
                else:
                    low = parse(subset.low).date() if subset.low is not None else None
                    high = (
                        parse(subset.high).date() if subset.high is not None else None
                    )
                if low is not None:
                    times = [time for time in times if time >= low]
                if high is not None:
                    times = [time for time in times if time <= high]
            elif isinstance(subset, Slice):
                point = parse(subset.point).date()
                times = [point]
        else:
            try:
                if isinstance(subset, Trim):
                    scaler.trim(dimension, subset.low, subset.high)
                elif isinstance(subset, Slice):
                    scaler.slice(dimension, subset.point)
            except WCSScalerUnknownDimension:
                raise WCS2Exception(
                    f"Invalid subsetting axis {subset.dimension}",
                    WCS2Exception.INVALID_AXIS_LABEL,
                    locator=subset.dimension,
                ) from None

    #
    # Transform spatial extent to native CRS.
    #
    scaler.to_crs(output_crs)

    #
    # Scaling
    #

    scales = request.scales
    if len(scales) != len({subset.axis.lower() for subset in scales}):
        axes = [subset.axis.lower() for subset in scales]
        duplicate_axes = [
            item for item, count in collections.Counter(axes).items() if count > 1
        ]
        raise WCS2Exception(
            f"Duplicate scales for ax{'i' if len(duplicate_axes) == 1 else 'e'}s: "
            f"{', '.join(duplicate_axes)}",
            WCS2Exception.INVALID_SCALE_FACTOR,
            locator=",".join(duplicate_axes),
        )

    for scale in scales:
        axis = scale.axis.lower()

        if axis in ("time", "k"):
            raise WCS2Exception(
                f"Cannot scale axis {scale.axis}",
                WCS2Exception.INVALID_SCALE_FACTOR,
                locator=scale.axis,
            )
        if isinstance(scale, ScaleAxis):
            scaler.scale_axis(axis, scale.factor)
        elif isinstance(scale, ScaleSize):
            scaler.scale_size(axis, scale.size)
        elif isinstance(scale, ScaleExtent):
            scaler.scale_extent(axis, scale.low, scale.high)

    #
    # Rangesubset
    #

    band_labels = layer.band_idx.band_labels(canonical=True)
    trans_band_labels = layer.band_idx.band_labels()
    if request.range_subset:
        bands = []
        for range_subset in request.range_subset:
            if isinstance(range_subset, str):
                if range_subset in band_labels:
                    bands.append(range_subset)
                elif range_subset in trans_band_labels:
                    bands.append(band_labels[trans_band_labels.index(range_subset)])
                else:
                    raise WCS2Exception(
                        f"No such field {range_subset}",
                        WCS2Exception.NO_SUCH_FIELD,
                        locator=range_subset,
                        valid_keys=band_labels,
                    )
            else:
                if (
                    range_subset.start not in band_labels
                    and range_subset.start not in trans_band_labels
                ):
                    raise WCS2Exception(
                        f"No such field {range_subset.start}",
                        WCS2Exception.ILLEGAL_FIELD_SEQUENCE,
                        locator=range_subset.start,
                        valid_keys=band_labels,
                    )
                if (
                    range_subset.end not in band_labels
                    and range_subset.end not in trans_band_labels
                ):
                    raise WCS2Exception(
                        f"No such field {range_subset.end}",
                        WCS2Exception.ILLEGAL_FIELD_SEQUENCE,
                        locator=range_subset.end,
                        valid_keys=band_labels,
                    )

                if range_subset.start in band_labels:
                    start = band_labels.index(range_subset.start)
                else:
                    start = trans_band_labels.index(range_subset.start)
                if range_subset.end in band_labels:
                    end = band_labels.index(range_subset.end)
                else:
                    end = trans_band_labels.index(range_subset.end)
                bands.extend(
                    band_labels[start : (end + 1) if end > start else (end - 1)]
                )
    # Uncomment to restore original styles parameter hack.
    #
    # elif styles:
    #     bands = get_bands_from_styles(styles, layer, version=2)
    #     if not bands:
    #         bands = band_labels
    else:
        bands = band_labels

    #
    # Format handling
    #

    if not request.format:
        fmt = cfg.wcs_formats_by_name[layer.native_format]  # type: ignore[index]
    else:
        try:
            fmt = cfg.wcs_formats_by_mime[request.format]
        except KeyError:
            raise WCS2Exception(
                f"Unsupported format: {request.format}",
                WCS2Exception.INVALID_PARAMETER_VALUE,
                locator="FORMAT",
                valid_keys=list(cfg.wcs_formats_by_mime),
            ) from None

    if len(times) > 1 and not fmt.multi_time:
        raise WCS2Exception(
            "Format does not support multi-time datasets - "
            "either constrain the time dimension or choose a different format",
            WCS2Exception.INVALID_SUBSETTING,
            locator="FORMAT or SUBSET",
        )
    affine = scaler.affine()
    geobox = GeoBox((scaler.size.y, scaler.size.x), affine, cfg.crs(output_crs))

    stacker = DataStacker(layer, geobox, times, bands=bands)
    qprof.end_event("setup")
    qprof.start_event("count-datasets")
    n_datasets = stacker.n_datasets()
    qprof.end_event("count-datasets")
    qprof["n_datasets"] = n_datasets

    try:
        layer.resource_limits.check_wcs(
            n_datasets,
            geobox.height,
            geobox.width,
            sum(layer.band_idx.dtype_size(b) for b in bands),
            len(times),
        )
    except ResourceLimited as e:
        if e.wcs_hard or not layer.low_res_product_names:
            raise WCS2Exception(
                f"This request processes too much data to be served in a reasonable amount of time. ({e}) "
                + "Please reduce the bounds of your request and try again."
            ) from None
        stacker.resource_limited = True
        qprof["resource_limited"] = str(e)

    if n_datasets == 0:
        raise WCS2Exception(
            "The requested spatio-temporal subsets return no data.",
            WCS2Exception.INVALID_SUBSETTING,
            http_response=404,
        )

    qprof.start_event("fetch-datasets")
    datasets = stacker.datasets()
    qprof.end_event("fetch-datasets")
    if qprof.active:
        qprof["datasets"] = {
            str(q): [str(i) for i in ids] for q, ids in stacker.dsids().items()
        }
    qprof.start_event("load-data")
    # FIXME: output can be None.
    output = stacker.data(datasets, skip_corrections=True)
    qprof.end_event("load-data")

    # Clean extent flag band from output
    raw_bands = [layer.band_idx.locale_band(b) for b in bands]
    for k, _ in output.data_vars.items():  # type: ignore[union-attr]
        if k not in raw_bands:
            output = output.drop_vars([k])  # type: ignore[union-attr]

    #
    # TODO: configurable
    #
    if fmt.mime == "image/geotiff":
        output = fmt.renderer(request.version)(
            request, output, output_crs, layer, scaler.size.x, scaler.size.y, affine
        )

    else:
        output = fmt.renderer(request.version)(request, output, output_crs)

    headers = {
        "Content-Type": fmt.mime,
        "content-disposition": f"attachment; filename={request.coverage_id}.{fmt.extension}",
    }
    headers.update(layer.resource_limits.wcs_cache_rules.cache_headers(n_datasets))
    return output, headers


def get_tiff(request, data, crs, product, width: int, height, affine):
    """Uses rasterio MemoryFiles in order to return a streamable GeoTiff response"""
    # Does not support multi-time dimension data - is this even possible in GeoTiff?
    supported_dtype_map = {
        "uint8": 1,
        "int8": 2,
        "uint16": 3,
        "int16": 4,
        "uint32": 5,
        "int32": 6,
        "float32": 7,
        "float64": 8,
        "complex": 10,
        "complex64": 11,
        "complex128": 12,
    }

    dtype_list = [data[array].dtype for array in data.data_vars]
    dtype = str(max(dtype_list, key=lambda d: supported_dtype_map[str(d)]))

    # TODO: convert other parameters as-well
    gtiff = request.geotiff_encoding_parameters
    cfg = get_config()

    if len(data.time) > 1:
        raise WCS2Exception("Multiple time slices not supported by GeoTIFF format")
    data = data.squeeze(dim="time", drop=True)
    data = data.astype(dtype)
    nodata = 0
    for band in data.data_vars:
        nodata = product.band_idx.nodata_val(band)
    with MemoryFile() as memfile:
        # pylint: disable=protected-access, bad-continuation

        kwargs = {}
        if gtiff.tile_width is not None:
            kwargs["blockxsize"] = gtiff.tile_width
        if gtiff.tile_height is not None:
            kwargs["blockysize"] = gtiff.tile_height

        if gtiff.predictor:
            predictor = gtiff.predictor.lower()
            if predictor == "horizontal":
                kwargs["predictor"] = 2
            elif predictor == "floatingpoint":
                kwargs["predictor"] = 3
        elif dtype == "float64":
            kwargs["predictor"] = 3
        else:
            kwargs["predictor"] = 2

        with memfile.open(
            driver="GTiff",
            width=width,
            height=height,
            count=len(data.data_vars),
            transform=affine,
            crs=crs,
            nodata=nodata,
            tiled=gtiff.tiling if gtiff.tiling is not None else True,
            compress=gtiff.compression.lower() if gtiff.compression else "lzw",
            interleave=gtiff.interleave or "band",
            dtype=dtype,
            **kwargs,
        ) as dst:
            for idx, band in enumerate(data.data_vars, start=1):
                dst.write(data[band].values, idx)
                dst.set_band_description(idx, product.band_idx.band_label(band))
                if cfg.wcs_tiff_statistics:
                    dst.update_tags(
                        idx, STATISTICS_MINIMUM=numpy.nanmin(data[band].values)
                    )
                    dst.update_tags(
                        idx, STATISTICS_MAXIMUM=numpy.nanmax(data[band].values)
                    )
                    dst.update_tags(
                        idx, STATISTICS_MEAN=numpy.nanmean(data[band].values)
                    )
                    dst.update_tags(
                        idx, STATISTICS_STDDEV=numpy.nanstd(data[band].values)
                    )
        return memfile.read()


def get_netcdf(request, data: xr.Dataset, crs) -> bytes:
    # Cleanup dataset attributes for NetCDF export
    data.attrs["crs"] = crs
    for v in data.data_vars.values():
        v.attrs["crs"] = crs
        if "spectral_definition" in v.attrs:
            del v.attrs["spectral_definition"]
        if "flags_definition" in v.attrs:
            del v.attrs["flags_definition"]
    if "time" in data and "units" in data["time"].attrs:
        del data["time"].attrs["units"]

    # And export to NetCDF
    return bytes(data.to_netcdf())



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import math
from collections.abc import Callable
from datetime import date, datetime, timezone
from typing import cast

import datacube
import odc.geo
import sqlalchemy.exc
from odc.geo.crs import CRS
from odc.geo.geom import Geometry
from psycopg.types.json import Json as Json3
from psycopg2.extras import Json as Json2
from sqlalchemy import text

from datacube_ows.index.api import CoordRange, LayerExtent, LayerSignature
from datacube_ows.index.postgres.mv_index import MVSelectOpts, mv_search
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_driver_name, get_sqlconn

_LOG: logging.Logger = logging.getLogger(__name__)


def jsonise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    if isinstance(bbox, dict):
        return bbox
    return {
        "top": bbox.top,
        "bottom": bbox.bottom,
        "left": bbox.left,
        "right": bbox.right,
    }


def create_range_entry(
    layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]
) -> None:
    # Take layer signature
    meta = LayerSignature(
        time_res=layer.time_resolution.value,
        products=tuple(layer.product_names),
        env=layer.local_env._name,
        datasets=layer.dc.index.datasets.count(product=layer.product_names),
    )

    conn = get_sqlconn(layer.dc)
    txn = conn.begin()
    if meta in cache:
        template = cache[meta][0]
        # A layer has same signature - reuse range"
        cache[meta].append(layer.name)
        try:
            conn.execute(
                text("""
            INSERT INTO ows.layer_ranges
                (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
            SELECT :layer_id, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated
            FROM ows.layer_ranges lr2
            WHERE lr2.layer = :template_id"""),
                {"layer_id": layer.name, "template_id": template},
            )
        except sqlalchemy.exc.IntegrityError:
            conn.execute(
                text("""
            UPDATE ows.layer_ranges lr1
            SET lat_min = lr2.lat_min,
                lat_max = lr2.lat_max,
                lon_min = lr2.lon_min,
                lon_max = lr2.lon_max,
                dates = lr2.dates,
                bboxes = lr2.bboxes,
                meta = lr2.meta,
                last_updated = lr2.last_updated
            FROM ows.layer_ranges lr2
            WHERE lr1.layer = :layer_id
            AND   lr2.layer = :template_id"""),
                {"layer_id": layer.name, "template_id": template},
            )
    else:
        Json = Json3 if get_driver_name(layer.dc.index) == "psycopg" else Json2
        # insert empty row if one does not already exist
        conn.execute(
            text("""
        INSERT INTO ows.layer_ranges
        (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
        VALUES
        (:p_layer, 0, 0, 0, 0, :empty, :empty, :meta, :now)
        ON CONFLICT (layer) DO NOTHING
        """),
            {
                "p_layer": layer.name,
                "empty": Json(""),
                "meta": Json(meta.as_json()),
                "now": datetime.now(tz=timezone.utc),
            },
        )

        prodids = [p.id for p in layer.products]
        # Update min/max lat/longs
        conn.execute(
            text(
                """
          UPDATE ows.layer_ranges lr
          SET lat_min = st_ymin(subq.bbox),
              lat_max = st_ymax(subq.bbox),
              lon_min = st_xmin(subq.bbox),
              lon_max = st_xmax(subq.bbox)
          FROM (
            SELECT st_extent(stv.spatial_extent) as bbox
            FROM ows.space_time_view stv
            WHERE stv.dataset_type_ref = ANY(:prodids)
          ) as subq
          WHERE lr.layer = :layer_id
          """
            ),
            {"layer_id": layer.name, "prodids": prodids},
        )

        # Set default timezone
        conn.execute(text("""set timezone to 'Etc/UTC'"""))

        # Loop over dates
        dates = set()  # Should get to here!
        if layer.time_resolution.is_solar():
            results = conn.execute(
                text(
                    """
              select
                    lower(temporal_extent), upper(temporal_extent),
                    ST_X(ST_Centroid(spatial_extent))
              from ows.space_time_view
              WHERE dataset_type_ref = ANY(:prodids)
              """
                ),
                {"prodids": prodids},
            )
            for result in results:
                dt1, dt2, lon = result
                dt = dt1 + (dt2 - dt1) / 2
                dt = dt.astimezone(timezone.utc)

                solar_day = datacube.api.query._convert_to_solar_time(dt, lon).date()
                dates.add(solar_day)
        else:
            results = conn.execute(
                text(
                    """
              select
                    array_agg(temporal_extent)
              from ows.space_time_view
              WHERE dataset_type_ref = ANY(:prodids)
              """
                ),
                {"prodids": prodids},
            )
            for result in results:
                for dat_ran in result[0]:
                    dates.add(dat_ran.lower)

        if layer.time_resolution.is_subday():
            date_formatter = lambda d: d.isoformat()  # noqa: E731
        else:
            date_formatter = lambda d: d.strftime("%Y-%m-%d")  # noqa: E731

        dates = sorted(dates)
        conn.execute(
            text("""
           UPDATE ows.layer_ranges
           SET dates = :dates
           WHERE layer= :layer_id
        """),
            {"dates": Json(list(map(date_formatter, dates))), "layer_id": layer.name},
        )
        # calculate bounding boxes
        # Get extent polygon from materialised views

        extent_4386 = cast(
            Geometry,
            mv_search(layer.dc.index, MVSelectOpts.EXTENT, products=layer.products),
        )

        all_bboxes = bbox_projections(extent_4386, layer.global_cfg.crses)

        conn.execute(
            text("""
        UPDATE ows.layer_ranges
        SET bboxes = :bbox
        WHERE layer = :layer_id
        """),
            {"bbox": Json(all_bboxes), "layer_id": layer.name},
        )

        cache[meta] = [layer.name]

    txn.commit()
    conn.close()


def bbox_projections(
    starting_box: odc.geo.Geometry, crses: dict[str, CRS]
) -> dict[str, dict[str, float]]:
    result = {}
    for crsid, crs in crses.items():
        if crs.valid_region is not None:
            clipped_crs_region = starting_box & crs.valid_region
            if clipped_crs_region.wkt == "POLYGON EMPTY":
                continue
            clipped_crs_bbox = clipped_crs_region.to_crs(crs).boundingbox
        else:
            clipped_crs_bbox = None
        if clipped_crs_bbox is not None:
            result[crsid] = jsonise_bbox(clipped_crs_bbox)
        else:
            projbbox = starting_box.to_crs(crs).boundingbox
            result[crsid] = sanitise_bbox(projbbox)
    return result


def sanitise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    def sanitise_coordinate(coord: float, fallback: float) -> float:
        return coord if math.isfinite(coord) else fallback

    return {
        "top": sanitise_coordinate(bbox.top, float("9.999999999e99")),
        "bottom": sanitise_coordinate(bbox.bottom, float("-9.999999999e99")),
        "left": sanitise_coordinate(bbox.left, float("-9.999999999e99")),
        "right": sanitise_coordinate(bbox.right, float("9.999999999e99")),
    }


def get_ranges(layer: OWSNamedLayer) -> LayerExtent | None:
    conn = get_sqlconn(layer.dc)
    results = conn.execute(
        text("""
        SELECT *
        FROM ows.layer_ranges
        WHERE layer=:pname"""),
        {"pname": layer.name},
    )

    for result in results:
        conn.close()
        if layer.time_resolution.is_subday():
            dt_parser: Callable[[str], datetime | date] = (  # noqa: E731
                lambda dts: datetime.fromisoformat(dts)
            )
        else:
            dt_parser = lambda dts: datetime.strptime(dts, "%Y-%m-%d").date()  # noqa: E731
        times = [dt_parser(d) for d in result.dates if d is not None]
        if not times:
            return None
        return LayerExtent(
            lat=CoordRange(min=float(result.lat_min), max=float(result.lat_max)),
            lon=CoordRange(min=float(result.lon_min), max=float(result.lon_max)),
            times=times,
            bboxes=result.bboxes,
        )
    return None
